# Standard library imports:
import logging
import numbers

# P3 imports:
from p3.base.string_helpers import safe_fmt
from p3.data.error import DataModuleError


_LOG = logging.getLogger("p3.data")


class Key(object):
    def __init__(self, val):
        assert isinstance(val, basestring) or isinstance(val, numbers.Integral)
        self._val = val

    def __repr__(self):
        return "Key({0:s})".format(repr(self._val))

    def val():
        def fget(self):
            return self._val
        return locals()
    val = property(**val())


class Operator(object):
    def __init__(self, *operands):
        self.ops = operands

    def __repr__(self):
        return "{0:s}( {1:s} )".format(
            self.__class__.__name__, ', '.join(repr(o) for o in self.ops))

    def get_op(self, op, row):
        if isinstance(op, Key):
            if op.val in row:
                return row[op.val]
            else:
                if isinstance(op.val, basestring):
                    code = DataModuleError.codes.unknown_column_name
                    msg  = safe_fmt("Unknown column name ('%s').", op.val)
                else:
                    code = DataModuleError.codes.invalid_column_index
                    msg  = safe_fmt("Invalid column index (%d).", op.val)
                raise DataModuleError(code, msg)
        else:
            return op


    @classmethod
    def is_key(cls, var):
        return isinstance(var, Key)


class Eq(Operator):
    def __init__(self, *operands):
        assert (len(operands) > 1)
        super(Eq, self).__init__(*operands)

    def test(self, row):
        op1 = self.get_op(self.ops[0], row)
        for op in self.ops[1:]:
            if self.get_op(op, row) != op1:
                return False
        return True


class Is(Operator):
    def __init__(self, *operands):
        assert (len(operands) > 1)
        super(Is, self).__init__(*operands)

    def test(self, row):
        op1 = self.get_op(self.ops[0], row)
        for op in self.ops[1:]:
            if self.get_op(op, row) is not op1:
                return False
        return True


class And(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self):
        return "{0:s}( {1:s} )".format(
            self.__class__.__name__, ', '.join(repr(a) for a in self.args))
    
    def test(self, row):
        for arg in self.args:
            if not arg.test(row):
                return False
        return True


class Not(object):
    def __init__(self, arg):
        self.arg = arg

    def test(self, row):
        return not self.arg.test(row)


class Or(object):
    def __init__(self, *args):
        self.args = args

    def __repr__(self):
        return "{0:s}({1:s})".format(
            self.__class__.__name__, ', '.join(repr(a) for a in self.args))

    def test(self, row):
        for arg in self.args:
            if arg.test(row):
                return True
        return False
