#
#  FILE NAME:
#      date_converter.py
#
#  AUTHOR(S):
#      Dariusz Walczak <dariusz.walczak@gmail.com>
#

# Standard library imports:
import datetime
import logging
import numbers
import re

# P3 imports:
from p3.base import Enum
from p3.logic.var.converter.var_converter import VarConverter


_LOG = logging.getLogger( "p3.logic" )


class DateConverter( VarConverter ):
    errors = Enum(
        "empty_input", "invalid_format", "invalid_date", "max_exceeded", "min_exceeded" )
    formats = Enum( "ddmmyy", "ddmmyyyy", "yymmdd", "yyyymmdd" )

    ## ===[ Initialization ]==================================================================== ##

    def __init__( self ):
        super( DateConverter, self ).__init__(
            self._deserialize, self._serialize, self.is_date_like )
        self.fmt = self.formats.yyyymmdd
        self.pattern = None
        self.max_val  = None
        self.min_val  = None


    def format( self, fmt ):
        assert fmt in self.formats

        self.fmt = fmt
        return self


    def max( self, val ):
        self.max_val = val
        return self


    def min( self, val ):
        self.min_val = val
        return self


    ## ===[ Conversion ]======================================================================== ##

    def _deserialize( self, res ):
        if not res.in_val:
            res.errors.append( self.errors.empty_input )
            return None

        if self.pattern is None:
            self.pattern = self._build_pattern( self.fmt )
        m = self.pattern.match( res.in_val )

        if m is not None:
            if m.group("sep1") != m.group("sep2"):
                res.errors.append( self.errors.invalid_format )
            else:
                if self.fmt in (self.formats.ddmmyy, self.formats.yymmdd):
                    y = 2000+int(m.group("year"))
                else:
                    y = int(m.group("year"))

                try:
                    out_val = datetime.date( y, int(m.group("month")), int(m.group("day")) )
                except ValueError as e:
                    res.errors.append( self.errors.invalid_date )
                else:
                    if (self.min_val is not None) and (out_val < self.min_val):
                        res.errors.append( self.errors.min_exceeded )
                    elif self.max_val is not None and out_val >= self.max_val:
                        res.errors.append( self.errors.max_exceeded )
        else:
            res.errors.append( self.errors.invalid_format )

        if not res.errors:
            return out_val
        else:
            return None

    @classmethod
    def is_date_like( self, obj ):
        return (hasattr(obj, 'year') and hasattr(obj, 'month') and hasattr(obj, 'day') and
                isinstance(obj.year, numbers.Integral) and
                isinstance(obj.month, numbers.Integral) and
                isinstance(obj.day, numbers.Integral))

    def _serialize( self, res ):
        try:
            datetime.date( res.in_val.year, res.in_val.month, res.in_val.day )
        except ValueError as e:
            res.errors.append( self.errors.invalid_date )
        else:
            fmt = { self.formats.ddmmyy:   "{d:02d}/{m:02d}/{y:02d}",
                    self.formats.ddmmyyyy: "{d:02d}/{m:02d}/{y:04d}",
                    self.formats.yymmdd:   "{y:02d}/{m:02d}/{d:02d}",
                    self.formats.yyyymmdd: "{y:04d}/{m:02d}/{d:02d}" }.get( self.fmt )
            assert fmt is not None

            if self.fmt in (self.formats.ddmmyy, self.formats.yymmdd):
                y = res.in_val.year % 100
            else:
                y = res.in_val.year

            out_val = fmt.format( y=y, m=res.in_val.month, d=res.in_val.day )

        if not res.errors:
            return out_val
        else:
            return None


    ## ===[ Internals ]========================================================================= ##

    def _build_pattern( self, fmt ):
        day_grp = "(?P<day>\d{2})"   # day
        mth_grp = "(?P<month>\d{2})" # month
        syr_grp = "(?P<year>\d{2})"  # short year
        lyr_grp = "(?P<year>\d{4})"  # long year
        sep     = "[-/.]{0,1}"       # separator (in future, allowed separators can depend on
                                     # converter configuration)
        sep1 = "(?P<sep1>%s)" % sep
        sep2 = "(?P<sep2>%s)" % sep

        if self.fmt == self.formats.ddmmyy:
            return re.compile( "^%s%s%s%s%s$" % (day_grp,sep1,mth_grp,sep2,syr_grp) )
        elif self.fmt == self.formats.ddmmyyyy:
            return re.compile( "^%s%s%s%s%s$" % (day_grp,sep1,mth_grp,sep2,lyr_grp) )
        elif self.fmt == self.formats.yymmdd:
            return re.compile( "^%s%s%s%s%s$" % (syr_grp,sep1,mth_grp,sep2,day_grp) )
        elif self.fmt == self.formats.yyyymmdd:
            return re.compile( "^%s%s%s%s%s$" % (lyr_grp,sep1,mth_grp,sep2,day_grp) )
